/**
//
// Author : Benoit Nabholz
// Last modification : Mars 2014
//
 
 
    This software is a computer program whose purpose is to compute population 
    genetic statistics out of ms simulation of ABC analysis used in the 
    study Nabholz et al. 2014 Transcriptome population genomics reveals 
    severe bottleneck and domestication cost in the African rice 
    (O. glaberrima)
     
    This software was created using Bio++ library :
    Copyright or © or Copr. Bio++ Development Team, (November 17, 2004)
 
    This software is governed by the CeCILL  license under French law and
    abiding by the rules of distribution of free software.  You can  use,
    modify and/ or redistribute the software under the terms of the CeCILL
    license as circulated by CEA, CNRS and INRIA at the following URL
    "http://www.cecill.info".
 
    As a counterpart to the access to the source code and  rights to copy,
    modify and redistribute granted by the license, users are provided only
    with a limited warranty  and the software's author,  the holder of the
    economic rights,  and the successive licensors  have only  limited
    liability.
 
    In this respect, the user's attention is drawn to the risks associated
    with loading,  using,  modifying and/or developing or reproducing the
    software by the user in light of its specific status of free software,
    that may mean  that it is complicated to manipulate,  and  that  also
    therefore means  that it is reserved for developers  and  experienced
    professionals having in-depth computer knowledge. Users are therefore
    encouraged to load and test the software's suitability as regards their
    requirements in conditions enabling the security of their systems and/or
    data to be ensured and,  more generally, to use and operate it in the
    same conditions as regards security.
 
    The fact that you are presently reading this means that you have had
    knowledge of the CeCILL license and that you accept its terms.


* 
**/


#include <Bpp/Seq/Io.all>
#include <Bpp/Seq/Alphabet.all>
#include <Bpp/Seq/Container.all>
#include <Bpp/Seq/Sequence.h>

#include <Bpp/PopGen/SequenceStatistics.h>
#include <Bpp/PopGen/PolymorphismSequenceContainerTools.h>
#include <Bpp/PopGen/PolymorphismSequenceContainer.h>

#include <Bpp/Utils.all>

#include <unistd.h>

using namespace std;
using namespace bpp;

/** **********************************************************************************************/
/** * From Fay and Wu 2000 Genetics 155: 1405–1413 ***********************************************/
/** * FayWu2000 is the ThetaH (eq. 3) in Fay and Wu 2000 *****************************************/
/** ancestralSites is a sequence containing the ancestral states in order to fold the mutation in the ingroup **/

double FayWu2000(const PolymorphismSequenceContainer& ingroup, Sequence* ancestralSites)
{
	if(ingroup.getNumberOfSites() != ancestralSites->size())
		throw Exception("ancestralSites and ingroup don't have the same size!!!'" );

	const Sequence& tmps = ingroup.getSequence(0);

	unsigned int alphabet_size = (ingroup.getAlphabet())->getSize();
	double value = 0.;
	for(unsigned int i = 0; i < ingroup.getNumberOfSites(); i++){

		const Site site = ingroup.getSite(i);
		string ancB = ancestralSites->getChar(i);
		int ancV = ancestralSites->getValue(i);
		
    if (! SiteTools::isConstant(site) || tmps.getChar(i) != ancB) {
			if(ancV < 0)
				continue;
			
      map<int, unsigned long int> count;
      SymbolListTools::getCounts(site, count);
      map<int, unsigned long int> tmp_k;
      unsigned int tmp_n = 0;
      for (map<int, unsigned long int>::iterator it = count.begin() ; it != count.end() ; it++){
        if (it->first >= 0 && it->first < static_cast<int>(alphabet_size)) {
					/** if derived allele **/
					if(it->first != ancV){
						tmp_k[it->first] = 2 * it->second * it->second;
					}
					tmp_n += it->second;
        }
			}
      if (tmp_n == 0 || tmp_n == 1) continue;
      for (map<int, unsigned long int>::iterator it = tmp_k.begin() ; it != tmp_k.end() ; it++){
        value += static_cast<double>(it->second) / (tmp_n * (tmp_n - 1));
			}
    }
  }
  return value;
}

/** **********************************************************************************************/
/**********************************************************************************               ****************/
/** Fst from eq. 3 of  Hudson, Slatkin and Maddison 1992 Genetics 132:153 ****************************************/
/********************************************************************************               ******************/
double FstHudson92(const PolymorphismSequenceContainer& psc, unsigned int id1, unsigned int id2, bool gapflag)
{
	vector<double> vdiff;
	double piIntra1, piIntra2, meanPiIntra, piInter, Fst;
	
	PolymorphismSequenceContainer *Pop1 = PolymorphismSequenceContainerTools::extractGroup(psc, id1);
	PolymorphismSequenceContainer *Pop2 = PolymorphismSequenceContainerTools::extractGroup(psc, id2);
	
  piIntra1 = SequenceStatistics::tajima83(*Pop1, false);
	piIntra2 = SequenceStatistics::tajima83(*Pop2, false);
	
	meanPiIntra = (piIntra1+piIntra2)/2;
	
	unsigned int N = 0;
	for(unsigned int i = 0; i<Pop1->getNumberOfSequences(); i++){
		const Sequence &s1 = Pop1->getSequence(i);
		for(unsigned int j = 0; j<Pop2->getNumberOfSequences(); j++){
			N++;
			const Sequence &s2 = Pop2->getSequence(j);
			vdiff.push_back(SiteContainerTools::computeSimilarity(s1,s2,true,"no gap",true));
		}
	}
	piInter = (VectorTools::sum(vdiff) / N)*psc.getNumberOfSites();
	
	
	Fst = 1.0 - meanPiIntra/piInter;
	
	delete Pop1;
	delete Pop2;
	
	return Fst;
	
}
/**********************************************************************************               ****************/
/** Nei et al. 1982 Fst =  (Pi Total - Pi intra) / Pi Total  ***************************************/
/********************************************************************************               ******************/

double FstNei82(const PolymorphismSequenceContainer& psc, unsigned int id1, unsigned int id2, bool gapflag)
{
	vector<double> vdiff;
	double piTotal, piIntra1, piIntra2, meanPiIntra, Fst;
	
	/** select ingroup sequences **/
	SequenceSelection ss;
	PolymorphismSequenceContainer* psci = dynamic_cast<PolymorphismSequenceContainer*>(psc.clone());
	for (unsigned int i = 0; i < psc.getNumberOfSequences(); i++){
		if (psc.getGroupId(i) != id1 && psc.getGroupId(i) != id2)
			ss.push_back(i);
	}
	for (unsigned int i = ss.size(); i > 0; i--){
		psci->getGroupId(i-1);
		psci->deleteSequence(ss[i - 1]);
	}
	piTotal = SequenceStatistics::tajima83(*psci, gapflag);

	
	PolymorphismSequenceContainer *Pop1 = PolymorphismSequenceContainerTools::extractGroup(psc, id1);
	PolymorphismSequenceContainer *Pop2 = PolymorphismSequenceContainerTools::extractGroup(psc, id2);
	
	piIntra1 = SequenceStatistics::tajima83(*Pop1, gapflag);
	piIntra2 = SequenceStatistics::tajima83(*Pop2, gapflag);
	
	
	meanPiIntra = (piIntra1+piIntra2)/2;
	
	Fst = 1 - meanPiIntra/piTotal;
	
	delete Pop1;
	delete Pop2;
	delete psci;	
	return Fst;
	
}

/**************************************************************************************************/
/** **********************************************************************************************/
/** compute number of sites where freq >0.5 in Pop1 and freq <0.5 in Pop2 ********************/

unsigned int jointFS(const PolymorphismSequenceContainer& psc, unsigned int id1, unsigned int id2, Sequence* ancestralSites)
{
	if(psc.getNumberOfSites() != ancestralSites->size())
		throw Exception("ancestralSites and ingroup don't have the same size!!!'" );

	PolymorphismSequenceContainer *Pop1 = PolymorphismSequenceContainerTools::extractGroup(psc, id1);
	PolymorphismSequenceContainer *Pop2 = PolymorphismSequenceContainerTools::extractGroup(psc, id2);

	unsigned int value = 0;
		
	for(unsigned int i = 0; i < Pop1->getNumberOfSites(); i++){

		const Site& site1 = Pop1->getSite(i);
		const Site& site2 = Pop2->getSite(i);
		
		string ancB = ancestralSites->getChar(i);
		int ancV = ancestralSites->getValue(i);
		
		if(ancV < 0)
			continue;
		
		map<int, double> freq1;
    SymbolListTools::getFrequencies(site1, freq1);
		
		map<int, double> freq2;
    SymbolListTools::getFrequencies(site2, freq2);
				
		int max1, max2;
		max1 = max2 = 0;
		double allele1, allele2;
		allele1 = allele2 = 0.0;
		
		
		/** found majority allele in group1 and group2 **/
		for (map<int, double>::iterator it = freq1.begin(); it != freq1.end(); it++){
			if (it->second > max1 && it->first != -1){
			max1 = it->second;
				allele1 = it->first;
			}
		}
		for (map<int, double>::iterator it = freq2.begin(); it != freq2.end(); it++){
			if (it->second > max2 && it->first != -1){
				max2 = it->second;
				allele2 = it->first;
			}
		}
		
		/** derived and freq > 0.5 only for the group1 **/
		if(max1 >= 0.5 && allele1 != ancV && max2 >= 0.8 && allele2 == ancV){
        value ++;
		}
  }
  delete Pop1;
	delete Pop2;
  return value;
}

/**************************************************************************************************/
/** **********************************************************************************************/
/** ********************************** Folded Frequencies spectrum           ********************/
vector<unsigned int> UnfoldedFreqSpectrum(const PolymorphismSequenceContainer& psc, Sequence* ancestralSites)
{
	if(psc.getNumberOfSites() != ancestralSites->size())
		throw Exception("ancestralSites and ingroup don't have the same size!!!'" );
	
	vector<unsigned int> FFS;
	for(unsigned int i = 0; i < psc.getNumberOfSites(); i++){

		const Site site = psc.getSite(i);
		
		if(i == 0){
			for(unsigned int j = 0; j < site.size()+1; j++){
				FFS.push_back(0);
			}
		}
		
		string ancB = ancestralSites->getChar(i);
		int ancV = ancestralSites->getValue(i);
		
		if(ancV < 0)
				continue;
			
		map<int, unsigned long int> count;
		SymbolListTools::getCounts(site, count);
		for (map<int, unsigned long int>::iterator it = count.begin() ; it != count.end() ; it++){
			/** if derived allele **/
			if(it->first != ancV){
				unsigned int pos = it->second;
				unsigned int c = FFS[pos];
				FFS[pos] = c+1;
				break;
			}else{
				unsigned int pos = site.size()-it->second;
				unsigned int c = FFS[pos];
				FFS[pos] = c+1;
				break;
			}
		}
	}
  return FFS;
}
/**************************************************************************************************/
string convertSeq(string seqN){
	string out;
	for(unsigned int i = 0; i<seqN.size(); i++){
		if(seqN[i] == '0')
			out.push_back('A');
		if(seqN[i] == '1')
			out.push_back('T');
	}
	return out;	
}
/********************************************************************/
/** PiInter : compute the Tajima's Pi between population ************/
/********************************************************************/

double PiInter(const PolymorphismSequenceContainer& psc, unsigned int id1, unsigned int id2)
{
	vector<double> vdiff;
	double piInter;
	
	PolymorphismSequenceContainer *Pop1 = PolymorphismSequenceContainerTools::extractGroup(psc, id1);
	PolymorphismSequenceContainer *Pop2 = PolymorphismSequenceContainerTools::extractGroup(psc, id2);
	
	
	unsigned int N = 0;
	for(unsigned int i = 0; i<Pop1->getNumberOfSequences(); i++){
		const Sequence &s1 = Pop1->getSequence(i);
		for(unsigned int j = 0; j<Pop2->getNumberOfSequences(); j++){
			N++;
			const Sequence &s2 = Pop2->getSequence(j);
			vdiff.push_back(SiteContainerTools::computeSimilarity(s1,s2,true,"no gap",true));
		}
	}
	piInter = (VectorTools::sum(vdiff) / N)*psc.getNumberOfSites();
	
	delete Pop1;
	delete Pop2;
	
	return piInter;
	
}
/**************************************************************************************************/
/************************************************************************************************/
/** 
Cree un tableau de contenant :
					Pop1
					Fixee ancestral	Polymorphisme	Fixee derive
			Fixee ancestral	Saa	Sap	Sad
Pop2	Polymorphisme	Spa	Spp	Spd
			Fixee derive	Sda	Sdp	Sdd
			
			Here derived is T and ancestral is A
**/

vector<unsigned int> polymorphismTable(const PolymorphismSequenceContainer& psc, unsigned int id1, unsigned int id2)
{
	
	unsigned int Saa, Sap, Sad, Spa, Spp, Spd, Sda, Sdp, Sdd;
	Saa = Sap = Sad = Spa = Spp = Spd = Sda = Sdp = Sdd = 0;
	vector<unsigned int> PT;
	
	PolymorphismSequenceContainer *Pop1 = PolymorphismSequenceContainerTools::extractGroup(psc, id1);
	PolymorphismSequenceContainer *Pop2 = PolymorphismSequenceContainerTools::extractGroup(psc, id2);
	
	for(unsigned int i = 0; i < Pop1->getNumberOfSites(); i++){

		const Site site1 = Pop1->getSite(i);
		const Site site2 = Pop2->getSite(i);
		
		/** define ancestral state (Here alway "A")**/
		int ancV = site1.getAlphabet()->charToInt("A");
		
		map<int, unsigned long int> count1, count2;
		SymbolListTools::getCounts(site1, count1);
		SymbolListTools::getCounts(site2, count2);
		
		
		if(SiteTools::isConstant(site1)){
			/** pop1 fixed ancetral **/
			if(count1[ancV] > 0){
				if(SiteTools::isConstant(site2)){
					if(count2[ancV] > 0)
						Saa++;
					if(count2[ancV] == 0)
						Sad++;
				}else{
						Sap++;
				}
			/** pop1 fixed derived **/
			}else{
				if(SiteTools::isConstant(site2)){
					if(count2[ancV] > 0)
						Sda++;
					if(count2[ancV] == 0)
						Sdd++;
				}else{
						Sdp++;
				}
			}
		}else{
		/** pop1 polymorphic **/
			if(SiteTools::isConstant(site2)){
				if(count2[ancV] > 0)
					Spa++;
				if(count2[ancV] == 0)
					Spd++;
			}else{
				Spp++;
			}
		}
	}
	PT.push_back(Saa);
	PT.push_back(Sap);
	PT.push_back(Sad);
	PT.push_back(Spa);
	PT.push_back(Spp);
	PT.push_back(Spd);
	PT.push_back(Sda);
	PT.push_back(Sdp);
	PT.push_back(Sdd);
	
	delete Pop1;
	delete Pop2;
	
	return PT;
}
/**************************************************************************************************/
void computeStats(VectorSiteContainer &seqCont, string selection, double bott, unsigned int nSimul){
	
	  double S_Sa ,S_Cu, W_Sa ,W_Cu , Pi_Sa ,Pi_Cu, Pi_Inter, FstNei, D_Sa ,D_Cu,  H_Sa, H_Cu;
// 		unsigned int jFS;
		vector< unsigned int> UFFS_Sa, UFFS_Cu;
		
		VectorSiteContainer *sitec = new VectorSiteContainer( seqCont );
		PolymorphismSequenceContainer * psc = new PolymorphismSequenceContainer( *sitec);
		
		for(unsigned int i = 0; i < psc->getNumberOfSequences(); i ++){
			string N = psc->getSequence(i).getName();
			StringTokenizer NST = StringTokenizer(N, "_");
			
			if(NST.getToken(0) == "in"){
				psc->setGroupId(i, 1);
				continue;
			}
			if(NST.getToken(0) == "out"){
				psc->setGroupId(i, 2);
				continue;
			}
		}
  
		PolymorphismSequenceContainer * pscCu = PolymorphismSequenceContainerTools::extractGroup (*psc, 1);
		PolymorphismSequenceContainer * pscSa = PolymorphismSequenceContainerTools::extractGroup (*psc, 2);
		
		if(psc->getNumberOfSites() == 0){
			cout << nSimul << "," << bott << "," << selection << ",0,0,0,0,0,0,0,NA,NA,NA,NA" << endl;
			delete sitec;
			delete pscSa;
			delete pscCu;
			delete psc;
			return;
		}
  
		/** ************************************************************** **/
		/** compute sequence statistic                                     **/
		/** ************************************************************** **/
		/** in the simulation A is ancestral and T is derived **/
		/** creat the ancestral site seq for compute the FayWu2000 **/
		const Sequence& seq = pscSa->getSequence(0);
		Sequence * ancestralSites = seq.clone();

		string acS;
		for(unsigned int i = 0; i < seq.size(); i++){
			acS.push_back('A');
		}
		ancestralSites->setContent(acS);
  
		S_Sa = SequenceStatistics::polymorphicSiteNumber(*pscSa, false );
		S_Cu = SequenceStatistics::polymorphicSiteNumber(*pscCu, false );
		W_Sa = SequenceStatistics::watterson75(*pscSa, false );
		W_Cu = SequenceStatistics::watterson75(*pscCu, false );
		Pi_Sa = SequenceStatistics::tajima83(*pscSa, false );
		Pi_Cu = SequenceStatistics::tajima83(*pscCu, false );
  

		double FWS = FayWu2000(*pscSa, ancestralSites);
		double FWC = FayWu2000(*pscCu, ancestralSites);
	
		H_Sa = Pi_Sa - FWS;
		H_Cu = Pi_Cu - FWC;
	
		Pi_Inter = PiInter(*psc, 1, 2);
		FstNei = FstNei82(*psc, 1, 2, false);
		
// 		vector<unsigned int> PT = polymorphismTable(*psc, 1, 2);
		
		if(S_Sa == 0){
			D_Sa = -999;
		}else{
			D_Sa = SequenceStatistics::tajimaDSS( *pscSa, false );
		}
  
		if(S_Cu == 0){
			D_Cu = -999;
		}else{
			D_Cu = SequenceStatistics::tajimaDSS( *pscCu, false );
		} 
		
// 		if(S_Sa == 0 && S_Cu == 0){
// 			Fst = -999.0;
// 		}else{
// 			Fst = FstHudson92(*psc, 1, 2, false);
// 		}
// 		jFS = jointFS(*psc, 1, 2, ancestralSites);
		
// 		UFFS_Sa = UnfoldedFreqSpectrum(*pscSa, ancestralSites);
// 		UFFS_Cu = UnfoldedFreqSpectrum(*pscCu, ancestralSites);
		
		/** only the no lost allele will be counted in the fraction of gene under selection **/
		cout << nSimul << "," << bott << "," << selection << ",";
		cout <<  S_Sa << "," << S_Cu << "," << W_Sa <<"," << W_Cu << ","; 
		cout << Pi_Sa <<"," << Pi_Cu << "," << Pi_Inter << "," << FstNei << ",";
		cout << D_Sa << "," << D_Cu << "," << H_Sa << "," << H_Cu << endl;
// 		for(unsigned int i = 0; i < pscCu->getNumberOfSequences()+1; i ++){
// 				cout << "," << UFFS_Sa[i];
// 		}
// 		for(unsigned int i = 0; i < pscCu->getNumberOfSequences()+1; i ++){
// 				cout << "," << UFFS_Cu[i];
// 		}
		
		
		delete sitec;
		delete pscSa;
		delete pscCu;
		delete psc;
}
/**************************************************************************************************/
int main (int args, char ** argv){
try{
    
if (args == 1)
{
  cout << "seq_stat [res_ms] [number_outgroup] [simulation_Ids]" << endl;
  return 0;
}

string nomfic = argv[1];
unsigned int noutgroup = TextTools::toInt(argv[2]);
string simulIds = argv[3];

/** print out seq **/
bool printSeq = false;

unsigned int nsam, i, c, segsites, nseq, simulId;
double bott;
string selection, line, seqname;
selection = "";

if(!FileTools::fileExists( nomfic )) {
  cerr << "ERROR!!! File " << nomfic << " does not exists." << endl;
  exit(-1);
}
ifstream resms (nomfic.c_str(), ios::in);

if(!FileTools::fileExists( simulIds )) {
  cerr << "ERROR!!! File " << simulIds << " does not exists." << endl;
  exit(-1);
}
ifstream FileSimulIds (simulIds.c_str(), ios::in);

const NucleicAlphabet * alpha = new DNA();
VectorSiteContainer * seqCont = new VectorSiteContainer(alpha);

c=nseq=0;
while(!resms.eof()){
	
	line = FileTools::getNextLine(resms);
	if(TextTools::isEmpty(line))
		continue;

	if(line[0] == '/' && line[1] == '/'){
		simulId = TextTools::toInt(FileTools::getNextLine(FileSimulIds));
		c++;
		
		StringTokenizer st = StringTokenizer(line, "\t");
		selection = st.getToken(1);
		bott = TextTools::toDouble(st.getToken(2));
		nsam = TextTools::toInt(st.getToken(3));
		
		if(selection == ""){
			cerr << "Probleme" << endl;
			return 1;
		}
		
		if(selection == "nosel"){
			StringTokenizer st = StringTokenizer(line);
			selection = "nosel";
			if(st.numberOfRemainingTokens() == 1){
				throw Exception("ms must be used with tbs option: // tbs in result file");
			}
// 			bott = TextTools::toDouble(st.getToken(2));
		}
		
		line = FileTools::getNextLine(resms);
		StringTokenizer st2 = StringTokenizer(line, ":");
		segsites = TextTools::toInt(TextTools::removeWhiteSpaces (st2.getToken(1)));
		
		if( segsites > 0 ){
			/** position line **/
			line = FileTools::getNextLine(resms);
			
			if(line[0] != 'p'){
				cout << line << " isn't position line!!!" << endl;
				return 0;
			}

			/** store seq **/
			for( i=0; i<nsam; i++) { 
				nseq++;
				if(nseq <= noutgroup){
					seqname = "out_"+TextTools::toString(nseq);
				}else{
					seqname = "in_"+TextTools::toString(nseq);
				}
				string seqN = FileTools::getNextLine(resms);
				string seqS = convertSeq(seqN);
				
				const Sequence& addseq = BasicSequence(seqname, seqS, alpha);
				seqCont->addSequence(addseq);
			}
			
			if(printSeq){
				string outFile = "seq."+TextTools::toString(c)+".fst";
				Fasta * FstSeq = new Fasta(1000000);
				FstSeq->writeAlignment(outFile , *seqCont);
				delete FstSeq;
			}
			
			computeStats(*seqCont, selection, bott, simulId);
			nseq = 0;
			seqCont->clear();
		}else{
			cout << simulId << "," << bott << "," << selection << ",0,0,0,0,0,0,0,0,NA,NA,NA,NA";
// 			for(unsigned int i = 0; i < 22; i ++){
// 				cout << ",0";
// 			}
			cout << endl;
		}
	}
}
// cout << c << endl;

delete alpha;
delete seqCont;

}
catch(exception & e){
  cout << e.what() << endl;
}
return 0;
}
